/* Sliding Tile Puzzle Solver */
/* Copyright C. A. Broadribb 1996 */
/* Completed 27/9/96 */

/* This program solves a sliding tile puzzle of eight numbered squares
   and one blank square.  The puzzle was devised by Sam Loyd. */

/* The constants MAXCOL and MAXROW are the number of columns and rows */
/* that the puzzle has.  They can be any number from 1 to 9.          */
/* This program uses the evaluation function f(s) = g(s) + h(s) where */
/* s is the state, g(s) is the number of moves taken to reach that    */
/* state from the initial state, and h(s) is the sum of the horizontal*/
/* and vertical distances of each tile from its correct position      */
/* (including the blank square).                                      */
/* This heuristic function is taken from the textbook Artificial      */
/* Intelligence - A Modern Approach by Russell and Norvig, p. 103.    */
/* It is the minimum number of moves required to solve the puzzle, so */
/* it will never over-estimate.                                       */

#include <stdio.h>
#include <stdlib.h>

#define MAXCOL 3
#define MAXROW 3
#define MAXNODE 1000
#define FALSE 0
#define TRUE 1

/* The array tiles contains the puzzle tiles - characters from 0  */
/* (which represents the blank square) to 9. The row and column   */
/* numbers of the blank square are stored (as blankrow and blankcol)  */
/* because they are frequently used.  hvalue is the result of the */
/* evaluation function performed on that configuration.           */

struct nodetype
{
	char tiles [3][3];
	int parentno, depth, hvalue, blankrow, blankcol, nodeno;
};

/* The combinations that the program has tried while searching    */
/* are stored here.  This is to allow the program to detect and   */
/* reject repeated combinations, so that it won't get stuck in an */
/* infinite loop. */ 

struct historytype
{
	struct nodetype node[MAXNODE];
	int historycount;  /* Points to next empty space */
};

/* This is the queue that is used to store nodes to search.  It stores */
/* the node numbers and also their evaluation function values (for     */
/* convenience only - if memory was an issue then they would not be    */
/* stored). */

struct queuetype
{
	int nodeno[MAXNODE];
	int hvalue[MAXNODE];
	int queuehead;  /* Points to next empty space */
};

void getconfig (struct nodetype *);
void getfinalstate (struct nodetype *);
void getinitstate (struct nodetype *, struct nodetype);
void printstate (struct nodetype);
void findposition (char, struct nodetype, int *, int *);
int countsteps (struct nodetype, struct nodetype);
void enqueue (struct nodetype, struct queuetype *);
void insertnode (struct nodetype, int, struct queuetype *);
void copynode (struct nodetype *, struct nodetype);
void archive (struct nodetype *, struct historytype *);
int getnode (struct nodetype *, struct queuetype *, struct historytype);
void expandnode (struct nodetype, struct queuetype *, struct historytype *, struct nodetype);
void expandit (struct nodetype, struct queuetype *, struct historytype *, struct nodetype, int, int);
int checkoriginality (struct nodetype, struct historytype);
int comparenode (struct nodetype, struct nodetype);
void displaypath (struct nodetype, struct historytype);
void cls ();

void main ()
{
/* The main loop is based upon the General Search Algorithm in the textbook */
/* Artificial Intelligence - A Modern Approach by Russell and Norvig. */

	int finished, failed;
	struct nodetype node;
	struct nodetype finalstate;
	struct historytype history;
	struct queuetype queue;

	history.historycount = 0;
	queue.queuehead = 0;

	getfinalstate (&finalstate);
	getinitstate (&node, finalstate);
	if (comparenode (node, finalstate))
		puts ("The puzzle is already solved!");
	else
	{
		clrscr();
		puts("Sliding Tile Puzzle Solver");
		puts("Copyright C. A. Broadribb 1996\n");
		puts ("I'm thinking...");
		archive (&node, &history);
		enqueue (node, &queue);
		finished = FALSE;
		failed = FALSE;
		do
		{
			if (getnode (&node, &queue, history) == FALSE)
				failed = TRUE;
			else
			{
				if (comparenode (node, finalstate))
				{
					displaypath (node, history);
					finished = TRUE;
				}
				else
					expandnode (node, &queue, &history, finalstate);
			}
		} while (!finished && !failed);
		if (failed)
			puts ("It is not possible to find a solution.");
		else
			puts ("Thanks for playing.  Have a nice day.");
	}
}

void getconfig (struct nodetype *node)
{
/* Ask the user to enter a puzzle configuration.  The user enters one */
/* row of numbers at a time.  0 represents the blank space. */

	int row, col, marker;
	char rowstring [MAXCOL];

	/* Initialise the configuration to all 0's */
	for (row = 0; row < MAXROW; row ++)
		for (col = 0; col < MAXCOL; col ++)
			node->tiles[row][col] = 0;

	printf ("The puzzle must have %i rows of %i columns.\n",MAXROW,MAXCOL);
	puts ("Enter each row of numbers (without any spaces).  Use 0 to represent");
	puts ("the blank square.");
	puts ("");
	for (row = 0; row < MAXROW; row ++)
	{
		printf ("Enter row #%i:  ",row + 1);
		gets (rowstring);
		marker = 0;
		for (col = 0; col < MAXCOL; col ++)
			node->tiles[row][col] = rowstring[marker++];
	}

}

void getfinalstate (struct nodetype *node)
{
/* Obtain the correct configuration of the puzzle from the user. */
/* This is only used to compare the current puzzle configuration */
/* against to see if the puzzle has been solved. */

	clrscr();
	puts("Sliding Tile Puzzle Solver");
	puts("Copyright C. A. Broadribb 1996\n");
	puts ("Enter the correct configuration of the puzzle.");
	getconfig (node);
}

void getinitstate (struct nodetype *node, struct nodetype finalstate)
{
/* Obtain the starting configuration of the puzzle from the user */
/* and initialise the appropriate data.  A parent number of -1   */
/* is used to indicate that it is the first node. */

	puts ("\nNow enter the starting configuration.");
	getconfig (node);
	node->parentno = -1;
	node->depth = 1;
	/* Initialise these two in case a blank square is not found. */
	node->blankrow = 0;
	node->blankcol = 0;
	findposition ('0',*node,&(node->blankrow),&(node->blankcol));
	node->hvalue = node->depth + countsteps (*node, finalstate);
}

void printstate (struct nodetype node)
{
/* Display the tile configuration for the node. Display the 0 as a blank */
/* space because it represents the blank square. */

	int row, col;

	for (row = 0; row < MAXROW; row ++)
	{
		for (col = 0; col < MAXCOL; col ++)
			if (node.tiles[row][col] == '0')
				printf(" ");
			else
				printf("%c",node.tiles[row][col]);
		puts ("");
	}
}

void findposition (char tile, struct nodetype node, int *tilerow, int *tilecol)
{
/* Search the node for the tile and returns its row and column if it */
/* is found.  Otherwise, tilerow and tilecol remain unchanged.       */

	int row, col;
	
	for (row = 0; row < MAXROW; row ++)
		for (col = 0; col < MAXCOL; col ++)
			if (node.tiles[row][col] == tile)
			{
				*tilerow = row;
				*tilecol = col;
			}
}

int countsteps (struct nodetype node, struct nodetype finalstate)
{
/* Sum the horizontal and vertical distances of every tile from  */
/* their correct positions given in finalstate and return the    */
/* result.  If a tile exists in node but not in finalstate or    */
/* vice versa then it is an error and is ignored. */

	int row, col, tilerow, tilecol, steps = 0;
	char tile;

	for (row = 0; row < MAXROW; row ++)
		for (col = 0; col < MAXCOL; col ++)
		{
			tile = finalstate.tiles[row][col];
			/* Initialise these two in case the tile is */
			/* not found in node. */
			tilerow = row;
			tilecol = col;
			findposition (tile,node,&tilerow,&tilecol);
			steps += abs(row-tilerow) + abs (col-tilecol);
		}
	return (steps);
}

void enqueue (struct nodetype node, struct queuetype *queue)
{
/* Place the node into the appropriate place in the queue depending */
/* on its hvalue.  The lowest hvalue is at the front of the queue.  */
/* The front of the queue is indicated by queue.queuehead.          */
/* archive() must be called before enqueue() so that the node       */
/* receives its node number. */

	char temp[10];

	int position = 0, doneit = FALSE;
	
	do
	{
		/* If the head of the queue has been reached then */
		/* place the node there and increment queuehead.  */
		if (position == queue->queuehead)
		{
			queue->hvalue[position] = node.hvalue;
			queue->nodeno[position] = node.nodeno;
			queue->queuehead ++;
			doneit = TRUE;
		}
		/* If the node's hvalue is higher than the one */
		/* at this position in the queue then insert the node here. */
		else if (node.hvalue > queue->hvalue[position])
		{
			insertnode (node,position,queue);
			doneit = TRUE;
		}
		position ++;
	} while (position <= queue->queuehead && doneit == FALSE);
}

void insertnode (struct nodetype node, int position, struct queuetype *queue)
{
/* Insert the node into the queue at the given position.  Shift all    */
/* the values in the nodeno and hvalue arrays to make room for the new */
/* data. This is not efficient but is easier to code than the          */
/* alternatives.  Increment queue.queuehead. */

	int loop;

	for (loop = queue->queuehead; loop > position; loop --)
	{
		queue->nodeno[loop] = queue->nodeno[loop-1];
		queue->hvalue[loop] = queue->hvalue[loop-1];
	}
	queue->nodeno[position] = node.nodeno;
	queue->hvalue[position] = node.hvalue;
	queue->queuehead ++;
}

void copynode (struct nodetype *dest, struct nodetype original)
{
/* Copy all of the original node's data to the destination node. */
	
	int row, col;

	for (row = 0; row < MAXROW; row ++)
		for (col = 0; col < MAXCOL; col ++)
			dest->tiles[row][col] = original.tiles[row][col];
	
	dest->parentno = original.parentno;
	dest->depth = original.depth;
	dest->hvalue = original.hvalue;
	dest->blankrow = original.blankrow;
	dest->blankcol = original.blankcol;
	dest->nodeno = original.nodeno;
}

void archive (struct nodetype *node, struct historytype *history)
{
/* Copy the node into the history structure so that it can be looked at  */
/* later when the program checks for repeated states.  Record the node's */
/* archive number in node.nodeno.  Increment the archive number          */
/* historycount) to point to the next available storage space.           */
/* If the program has run out of memory (ie historycount is too large)   */
/* then it will detect it and exit. */

	int counter;

	if (history->historycount+1 < MAXNODE)
	{       
		node->nodeno = history->historycount;
		copynode (&(history->node[history->historycount]),*node);
		history->historycount ++;
	}
	else
	{
		puts ("I've run out of memory!");
		exit(0);
	}
}

int getnode (struct nodetype *node, struct queuetype *queue, struct historytype history) 
{
/* If the queue is empty then return false.  Otherwise, */
/* remove an entry from the queue and decrement queuehead.  Find the  */
/* relevant entry in history and copy its data to node.  Return true. */

	int anode, gotit = FALSE;

	if (queue->queuehead > 0)
	{
		queue->queuehead --;
		anode = queue->nodeno[queue->queuehead];
		copynode (node,history.node[anode]);
		gotit = TRUE;   
	}
	return (gotit);
}

void expandnode(struct nodetype node, struct queuetype *queue, struct historytype *history, struct nodetype finalstate) 
{
/* Produce child nodes of node by moving the blank square up, down, */
/* left and right. */

	expandit (node,queue,history,finalstate,node.blankrow-1,node.blankcol);
	expandit (node,queue,history,finalstate,node.blankrow+1,node.blankcol);
	expandit (node,queue,history,finalstate,node.blankrow,node.blankcol-1);
	expandit (node,queue,history,finalstate,node.blankrow,node.blankcol+1);
}

void expandit (struct nodetype parentnode, struct queuetype *queue, struct historytype *history, struct nodetype finalstate, int row, int col) 
{
/* Check whether the row and column are valid.  If so, create a new node */
/* with the blank square swapped with the tile in the row and column.    */
/* Check if the result is a configuration that has not yet been examined.*/
/* If so, place it into history and the queue. */
/* If these criteria aren't met then do nothing. */

	struct nodetype tempnode;

	if (col >= 0 && col < MAXCOL && row >= 0 && row < MAXROW)
	{
		copynode (&tempnode,parentnode);
		/* Swap the blank square with the square at row, col */
		tempnode.tiles[tempnode.blankrow][tempnode.blankcol] = tempnode.tiles[row][col];
		tempnode.tiles[row][col] = '0';
		tempnode.blankrow = row;
		tempnode.blankcol = col;        
		/* Check to make sure that it's not a repeated configuration */
		if (checkoriginality (tempnode,*history))
		{
			tempnode.depth = parentnode.depth + 1;
			tempnode.parentno = parentnode.nodeno;
			tempnode.hvalue = tempnode.depth + countsteps (tempnode, finalstate);
			archive (&tempnode,history);
			enqueue (tempnode,queue);
		}
	}
}

int checkoriginality (struct nodetype node, struct historytype history)
{
/* Search through history checking to see whether node is in there. */
/* If the node is an original configuration then return true, else    */
/* return false. */

	int position = 0, original = TRUE;
	
	while (original == TRUE && position < history.historycount)
	{
		if (comparenode (history.node[position],node))
			original = FALSE;
		position ++;
	}
	return (original);
}

int comparenode (struct nodetype node1, struct nodetype node2)
{
/* Compare node1 with node2 and return true if the tile configuration */
/* in both is the same.  Return false if they are different.          */
	
	int same, row, col;

	same = TRUE;
	for (row = 0; row < MAXROW; row ++)
		for (col = 0; col < MAXCOL; col ++)
			if (node1.tiles[row][col] != node2.tiles[row][col])
				same = FALSE;
	return (same);
}

void displaypath (struct nodetype node, struct historytype history)
{
/* Search back through history, starting at the current node, until   */
/* the original node is reached.  Record the steps in the array path. */
/* Then display the path, starting from the original node.            */
/* This will display the steps to solve the puzzle, from the original */
/* configuration to the final configuration. */

	int path[100];
	int loop, currentno, pathend = 0;
	char temp [10];

	currentno = node.nodeno;
	do
	{
		path[pathend] = currentno;
		pathend ++;
		currentno = history.node[currentno].parentno;
	} while (currentno >= 0);
	for (loop = pathend-1; loop >= 0; loop --)
	{
		clrscr();
		puts ("These are the steps to solve the puzzle:");
		puts ("");
		printstate (history.node[path[loop]]);
		puts ("");
		puts ("Press enter to continue.");
		gets (temp);
	}
}

